import importlib.util
import os
import re
import threading
import yaml

from builtins import str
from distutils.spawn import find_executable
from dotmap import DotMap

from popper.cli import log


def setup_base_cache():
    """Set up the base cache directory.

    Args:
      None

    Returns:
      str: The path to the base cache directory.
    """
    if os.environ.get('POPPER_CACHE_DIR', None):
        base_cache = os.environ['POPPER_CACHE_DIR']
    else:
        cache_dir = os.environ.get('XDG_CACHE_HOME',
                                   os.path.join(os.environ['HOME'], '.cache'))
        base_cache = os.path.join(cache_dir, 'popper')

    os.makedirs(base_cache, exist_ok=True)

    return base_cache


def decode(line):
    """Make treatment of stdout Python 2/3 compatible.

    Args:
      line(str): The string that is required to be converted.

    Returns:
      str : The string in converted form.
    """
    if isinstance(line, bytes):
        return line.decode('utf-8')
    return line


class threadsafe_iter_3:
    """Takes an iterator/generator and makes it thread-safe by serializing call
    to the `next` method of given iterator/generator."""

    def __init__(self, it):
        self.it = it
        self.lock = threading.Lock()

    def __iter__(self):
        return self

    def __next__(self):
        with self.lock:
            return self.it.__next__()


def threadsafe_generator(f):
    """A decorator that takes a generator function and makes it thread-safe.

    Args:
      f(function): Generator function

    Returns:
      None
    """
    def g(*args, **kwargs):
        """

        Args:
          *args(list): List of non-key worded,variable length arguments.
          **kwargs(dict): List of key-worded,variable length arguments.

        Returns:
          function: The thread-safe function.
        """
        return threadsafe_iter_3(f(*args, **kwargs))
    return g


def sanitized_name(name, wid=''):
    """Clean a step name and change it to proper format. It replaces all the
    unwanted characters with `_`.

    Args:
      name(str): The crud step name.
      wid(str): It  is a workflow ID produced by a utils.get_id().

    Returns:
      str: The sanitize step name.
    """
    return f"popper_{re.sub('[^a-zA-Z0-9_.-]', '_', name)}_{wid}"


def of_type(param, valid_types):
    """Function to check the type of a parameter.

    It tries to match the type of the parameter with the
    types passed through `valid_types` list.

    Args:
      param: A value of any type.
      valid_types(list): A list of acceptable types.

    Returns:
      bool: True/False, depending upon whether the type of
      the passed param matches with any of the valid types.
    """
    for t in valid_types:
        if t == 'str':
            if isinstance(param, str):
                return True

        if t == 'dict':
            if isinstance(param, dict):
                return True

        if t == 'los':
            if isinstance(param, list):
                res = list(map(lambda a: isinstance(a, str), param))
                return False not in res

    return False


def write_file(path, content=''):
    """Create and write contents to a file. If no content is provided a blank
    file is created.

    Args:
      path(str): The path where the file would be created.
      content(str, optional): The content to write in the file.
                            (Default value = '')

    Returns:
      None
    """
    f = open(path, 'w')
    f.write(content)
    f.close()


def module_from_file(module_name, file_path):
    """Import a file as a module.

    Args:
      module_name(str): The name of the module.
      file_path(str): The path to the file to be imported.

    Returns:
      module
    """
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def load_config_file(config_file):
    """Validate and parse the engine configuration file.

    Args:
      config_file(str): Path to the file to be parsed.

    Returns:
      dict: Engine configuration.
    """
    if not os.path.exists(config_file):
        log.fail(f'File {config_file} was not found.')

    if not config_file.endswith('.py'):
        log.fail('Configuration file must be a python source file.')

    module_name = os.path.basename(config_file)[:-3]
    module = module_from_file(module_name, config_file)

    return module


def assert_executable_exists(command):
    """Check if the given command can be invoked; fails if not."""
    if not find_executable(command):
        log.fail(f"Could not find '{command}'.")


def prettystr(a):
    if type(a) == DotMap:
        a = a.toDict()
    if type(a) == os._Environ:
        a = dict(a)
    if type(a) == dict:
        return f'{yaml.dump(a, default_flow_style=False)}'



from __future__ import unicode_literals

import re
import hcl
import os
import yaml

from copy import deepcopy
from builtins import str, dict
from popper.cli import log as log

import popper.utils as pu


VALID_STEP_ATTRS = ["uses", "args", "needs", "runs", "secrets", "env", "name", "next"]


class Workflow(object):
    """Represents an immutable workflow."""

    def __init__(self):
        pass

    def parse(self, substitutions=None, allow_loose=False):
        """Parse and validate a workflow.

        Args:
          substitutions(list): Substitutions that are to be passed
                                as an arguments. (Default value = None)
          allow_loose(bool): Flag if the unused variables are to be
                                ignored. (Default value = False)

        Returns:
            None.

        """
        self.validate_workflow_block()
        self.validate_step_blocks()
        self.normalize()
        self.parse_substitutions(substitutions, allow_loose)
        self.check_for_broken_workflow()
        self.complete_graph()

    def complete_graph(self):
        raise NotImplementedError(
            "This method is required to be implemented in the derived class."
        )

    def normalize(self):
        raise NotImplementedError(
            "This method is required to be implemented in the derived class."
        )

    @staticmethod
    def new(file, step=None, skipped_steps=[], substitutions=[],
            allow_loose=False, include_step_dependencies=False):
        if file.endswith('.workflow'):
            wf = HCLWorkflow(file)
        elif file.endswith('.yml') or file.endswith('.yaml'):
            wf = YMLWorkflow(file)
        else:
            log.fail('Unrecognized workflow file format.')

        wf.parse(substitutions=substitutions, allow_loose=allow_loose)

        log.debug(
            f'Parsed workflow:\n'
            f'{yaml.dump(wf, default_flow_style=False, default_style="")}')

        wf = Workflow.skip_steps(wf, skipped_steps)
        wf = Workflow.filter_step(wf, step, include_step_dependencies)

        wf.check_for_unreachable_steps(skipped_steps)

        return wf

    @staticmethod
    def format_command(params):
        """A static method that formats the `runs` and `args` attributes into a
        list of strings.

        Args:
          params(list/str): run or args that are being executed.

        Returns:
            list: List of strings of parameters.
        """
        if pu.of_type(params, ['str']):
            return params.split(" ")
        return params

    @pu.threadsafe_generator
    def get_stages(self):
        """Generator of stages. A stages is a list of steps that can be
        executed in parallel.
        """
        def resolve_intersections(stage):
            """Removes steps from a stage that creates conflict between the
            selected stage candidates.

            Args:
              stage(set): Stage from which conflicted steps are to be
                            removed.

            Returns:
                None
            """
            steps_to_remove = set()
            for a in stage:
                if self.steps[a].get('next', None):
                    intersection = self.steps[a]['next'].intersection(stage)
                    if intersection:
                        for i in intersection:
                            steps_to_remove.add(i)

            for a in steps_to_remove:
                stage.remove(a)

        current_stage = self.root

        while current_stage:
            yield current_stage
            next_stage = set()
            for n in current_stage:
                next_stage.update(self.steps[n].get('next', set()))
                resolve_intersections(next_stage)
            current_stage = next_stage

    def check_for_broken_workflow(self):
        step_dependencies = set()
        for a_name, a_block in self.steps.items():
            step_dependencies.update(set(a_block.get('needs', list())))

        if self.wf_fmt == "hcl":
            step_dependencies.update(set(self.resolves))

        for step in step_dependencies:
            if step not in self.steps:
                log.fail(f"Step '{step}' referenced in workflow but missing")

    def validate_workflow_block(self):
        """Validate the syntax of the workflow block."""
        if self.wf_fmt == 'yml':
            return

        workflow_block_cnt = len(
            self.wf_dict.get(
                'workflow', dict()).items())
        if workflow_block_cnt == 0:
            log.fail('A workflow block must be present.')

        if workflow_block_cnt > 1:
            log.fail('Cannot have more than one workflow blocks.')

        workflow_block = list(self.wf_dict['workflow'].values())[0]
        for key in workflow_block.keys():
            if key not in VALID_WORKFLOW_ATTRS:
                log.fail(f"Invalid workflow attribute '{key}'.")

        if not workflow_block.get('resolves', None):
            log.fail('[resolves] attribute must be present in a '
                     'workflow block.')

        if not pu.of_type(workflow_block['resolves'], ['str', 'los']):
            log.fail('[resolves] attribute must be a string or a list '
                     'of strings.')

        if workflow_block.get('on', None):
            if not pu.of_type(workflow_block['on'], ['str']):
                log.fail('[on] attribute mist be a string.')

    def validate_step_blocks(self):
        """Validate the syntax of the step blocks."""
        if not self.wf_dict.get('steps', None):
            log.fail('At least one step block must be present.')

        for _, a_block in self.wf_dict['steps'].items():
            for key in a_block.keys():
                if key not in VALID_STEP_ATTRS:
                    log.fail(f"Invalid step attribute '{key}'.")

            if not a_block.get('uses', None):
                log.fail('[uses] attribute must be present in step block.')

            if not pu.of_type(a_block['uses'], ['str']):
                log.fail('[uses] attribute must be a string.')

            if a_block.get('needs', None):
                if not pu.of_type(a_block['needs'], ['str', 'los']):
                    log.fail(
                        '[needs] attribute must be a string or a list '
                        'of strings.')

            if a_block.get('args', None):
                if not pu.of_type(a_block['args'], ['str', 'los']):
                    log.fail(
                        '[args] attribute must be a string or a list '
                        'of strings.')

            if a_block.get('runs', None):
                if not pu.of_type(a_block['runs'], ['str', 'los']):
                    log.fail(
                        '[runs] attribute must be a string or a list '
                        'of strings.')

            if a_block.get('env', None):
                if not pu.of_type(a_block['env'], ['dict']):
                    log.fail('[env] attribute must be a dict.')

            if a_block.get('secrets', None):
                if not pu.of_type(a_block['secrets'], ['str', 'los']):
                    log.fail(
                        '[secrets] attribute must be a string or a list '
                        'of strings.')

    def check_for_unreachable_steps(self, skip=None):
        """Validates a workflow by checking for unreachable nodes / gaps in the
        workflow.

        Args:
          skip(list, optional): The list of steps to skip if applicable.
                                (Default value = None)

        Returns:
            None
        """
        if not skip or self.wf_fmt == 'yml':
            # noop
            return

        def _traverse(entrypoint, reachable, steps):
            """

            Args:
              entrypoint(set): Set containing the entry point of part of the
                                workflow.
              reachable(set): Set containing all the reachable parts of
                                workflow.
              steps(dict): Dictionary containing the identifier of the
                                workflow and its description.

            Returns:
                None
            """
            for node in entrypoint:
                reachable.add(node)
                _traverse(steps[node].get(
                    'next', []), reachable, steps)

        reachable = set()
        skipped = set(self.props.get('skip_list', []))
        steps = set(map(lambda a: a[0], self.steps.items()))

        _traverse(self.root, reachable, self.steps)

        unreachable = steps - reachable
        if unreachable - skipped:
            if skip:
                log.fail(
                    f'Unreachable step(s): {", ".join(unreachable-skipped)}.')
            else:
                log.warning(f'Unreachable step(s): {", ".join(unreachable)}.')

        for a in unreachable:
            self.steps.pop(a)

    def parse_substitutions(self, substitutions=None, allow_loose=False):
        """

        Args:
          substitutions(list): List of substitutions that are passed
                                as an arguments.
          allow_loose(bool): Flag used to ignore unused substitution
                                variable in the workflow.

        Returns:
            None

        """

        if not substitutions:
            # noop
            return

        substitution_dict = dict()

        for args in substitutions:
            item = args.split('=', 1)
            if len(item) < 2:
                raise Exception("Excepting '=' as seperator")
            substitution_dict['$' + item[0]] = item[1]

        for keys in substitution_dict:
            if not re.match(r"\$_[A-Z0-9]+", keys):
                log.fail(f"Improper substitution format: '{keys}'.")

        used = {}

        for wf_name, wf_block in self.steps.items():

            attr = wf_block.get('needs', [])
            for i in range(len(attr)):
                for k, v in substitution_dict.items():
                    if k in attr[i]:
                        used[k] = 1
                        attr[i] = attr[i].replace(k, v)

            attr = wf_block.get('uses', '')
            for k, v in substitution_dict.items():
                if k in attr:
                    used[k] = 1
                    wf_block['uses'] = attr.replace(k, v)

            attr = wf_block.get('args', [])
            for i in range(len(attr)):
                for k, v in substitution_dict.items():
                    if k in attr[i]:
                        used[k] = 1
                        attr[i] = attr[i].replace(k, v)

            attr = wf_block.get('runs', [])
            for i in range(len(attr)):
                for k, v in substitution_dict.items():
                    if k in attr[i]:
                        used[k] = 1
                        attr[i] = attr[i].replace(k, v)

            attr = wf_block.get('secrets', [])
            for i in range(len(attr)):
                for k, v in substitution_dict.items():
                    if k in attr[i]:
                        used[k] = 1
                        attr[i] = attr[i].replace(k, v)

            attr = wf_block.get('env', {})
            temp_dict = {}
            for key in attr.keys():
                check_replacement = False
                for k, v in substitution_dict.items():
                    if k in key:
                        used[k] = 1
                        temp_dict[v] = attr[key]
                        check_replacement = True

                if(check_replacement is False):
                    temp_dict[key] = attr[key]

            for key, value in temp_dict.items():
                for k, v in substitution_dict.items():
                    if k in value:
                        used[k] = 1
                        temp_dict[key] = v

            if(len(temp_dict) != 0):
                wf_block['env'] = temp_dict

        if not allow_loose and len(substitution_dict) != len(used):
            log.fail("Not all given substitutions are used in"
                     "the workflow file")

    @staticmethod
    def skip_steps(wf, skip_list=[]):
        """Removes the steps to be skipped from the workflow graph and return
        a new `Workflow` object.

        Args:
          wf(Workflow): The workflow object to operate upon.
          skip_list(list): List of steps to be skipped.
                            (Default value = list())

        Returns:
          Workflow : The updated workflow object.
        """
        if not skip_list:
            # noop
            return wf

        workflow = deepcopy(wf)
        for sa_name in skip_list:
            if sa_name not in workflow.steps:
                log.fail(f"Referenced step '{sa_name} missing.")
            sa_block = workflow.steps[sa_name]
            # Clear up all connections from sa_block
            sa_block.get('next', set()).clear()
            del sa_block.get('needs', list())[:]

            # Handle skipping of root step's
            if sa_name in workflow.root:
                workflow.root.remove(sa_name)

            # Handle skipping of non-root step's
            for a_name, a_block in workflow.steps.items():
                if sa_name in a_block.get('next', set()):
                    a_block['next'].remove(sa_name)

                if sa_name in a_block.get('needs', list()):
                    a_block['needs'].remove(sa_name)

        workflow.props['skip_list'] = list(skip_list)
        return workflow

    @staticmethod
    def filter_step(wf, step, with_dependencies=False):
        """Filters out all steps except the one passed in the argument from
        the workflow.
        Args:
          wf(Workflow): The workflow object to operate upon.
          step(str): The step to run.
          with_dependencies(bool, optional): Filter out step to
        run with dependencies or not. (Default value = False)
        Returns:
          Workflow: The updated workflow object.
        """

        if not step:
            # noop
            return wf

        # Recursively generate root when an step is run
        # with the `--with-dependencies` flag.
        def find_root_recursively(workflow, step, required_steps):
            """
            Args:
              workflow(worklfow): The workflow object to operate upon.
              step(str): The step to run.
              required_steps(set): Set containing steps that are
                                    to be executed.
            Returns:
                None
            """
            required_steps.add(step)
            if workflow.steps[step].get('needs', None):
                for a in workflow.steps[step]['needs']:
                    find_root_recursively(workflow, a, required_steps)
                    if not workflow.steps[a].get('next', None):
                        workflow.steps[a]['next'] = set()
                    workflow.steps[a]['next'].add(step)
            else:
                workflow.root.add(step)

        # The list of steps that needs to be preserved.
        workflow = deepcopy(wf)

        if step not in workflow.steps:
            log.fail(f"Referenced step '{step}' missing.")

        steps = set(map(lambda x: x[0], workflow.steps.items()))

        required_steps = set()

        if with_dependencies:
            # Prepare the graph for running only the given step
            # only with its dependencies.
            find_root_recursively(workflow, step, required_steps)

            filtered_steps = steps - required_steps

            for ra in required_steps:
                a_block = workflow.steps[ra]
                common_steps = filtered_steps.intersection(
                    a_block.get('next', set()))
                if common_steps:
                    for ca in common_steps:
                        a_block['next'].remove(ca)
        else:
            # Prepare the step for its execution only.
            required_steps.add(step)

            if workflow.steps[step].get('next', None):
                workflow.steps[step]['next'] = set()

            if workflow.steps[step].get('needs', None):
                workflow.steps[step]['needs'] = list()

            workflow.root.add(step)

        # Make the list of the steps to be removed.
        steps = steps - required_steps

        # Remove the remaining steps
        for a in steps:
            if a in workflow.root:
                workflow.root.remove(a)
            workflow.steps.pop(a)

        return workflow


class YMLWorkflow(Workflow):
    """Parse a yml based workflow and generate the workflow graph.
    """

    def __init__(self, wfile):
        """Loads the workflow as a dict from the `.yml` file."""
        super(YMLWorkflow, self).__init__()
        self.wf_fmt = "yml"

        if not os.path.exists(wfile):
            # try to load string
            self.wf_list = yaml.safe_load(wfile)['steps']
        else:
            with open(wfile) as fp:
                self.wf_list = yaml.safe_load(fp)['steps']
                if not self.wf_list:
                    return

        self.wf_dict = {'steps': dict()}
        self.id_map = dict()

        for idx, step in enumerate(self.wf_list):
            # If no id attribute present, make one
            _id = str(step.get('id', idx + 1))
            self.wf_dict['steps'][_id] = step
            self.id_map[idx + 1] = _id
            step.pop('id', None)

        self.name = os.path.basename(wfile)[:-4]

    def normalize(self):
        """Takes properties from the `self.wf_dict` dict and makes them
        native to the `Workflow` class. Also it normalizes some of the
        attributes of a parsed workflow according to the Github defined
        specifications.

        For example, it changes `args`, `runs` and `secrets` attribute,
        if provided as a string to a list of string by splitting around
        whitespace. Also, it changes parameters like `uses` and `resolves`,
        if provided as a string to a list.

        Args:
            None

        Returns:
            None
        """
        self.on = ""
        self.root = set()
        self.props = dict()
        self.steps = self.wf_dict['steps']

        for a_name, a_block in self.steps.items():
            a_block['name'] = a_name

            if a_block.get('needs', None):
                if pu.of_type(a_block['needs'], ['str']):
                    a_block['needs'] = [a_block['needs']]

            if a_block.get('args', None):
                a_block['args'] = Workflow.format_command(a_block['args'])

            if a_block.get('runs', None):
                a_block['runs'] = Workflow.format_command(a_block['runs'])

            if a_block.get('secrets', None):
                a_block['secrets'] = Workflow.format_command(
                    a_block['secrets'])

    def get_containing_set(self, idx):
        """Find the set from the list of sets of step dependencies, where
        the step with the given index is present.

        Args:
            idx(int): The index of the step to be searched.

        Returns:
            set: The required set.
        """
        for _set in self.dep_sets:
            if self.id_map[idx] in _set:
                return _set

        required_set = set()
        required_set.add(self.id_map[idx])
        return required_set

    def complete_graph(self):
        """Function to generate the workflow graph by adding forward edges.

        Args:
            None

        Returns:
            None
        """
        # Connect the graph as much as possible.
        for a_id, a_block in self.steps.items():
            if a_block.get('needs', None):
                for a in a_block['needs']:
                    if not self.steps[a].get('next', None):
                        self.steps[a]['next'] = set()
                    self.steps[a]['next'].add(a_id)

        # Generate the dependency sets.
        self.dep_sets = list()
        self.visited = dict()

        for a_id, a_block in self.steps.items():
            if a_block.get('next', None):
                if a_block['next'] not in self.dep_sets:
                    self.dep_sets.append(a_block['next'])
                    self.visited[tuple(a_block['next'])] = False

            if a_block.get('needs', None):
                if a_block['needs'] not in self.dep_sets:
                    self.dep_sets.append(set(a_block['needs']))
                    self.visited[tuple(a_block['needs'])] = False

        # Moving from top to bottom
        for idx, id in self.id_map.items():
            step = self.steps[id]
            if not step.get('next', None):
                # if this is not the last step,
                if idx + 1 <= len(self.steps):
                    curr = self.id_map[idx]
                    next = self.id_map[idx + 1]
                    # If the current step and next step is not in any
                    # set,
                    if ({curr, next} not in self.dep_sets) and (
                            {next, curr} not in self.dep_sets):
                        next_set = self.get_containing_set(idx + 1)
                        curr_set = self.get_containing_set(idx)

                        if not self.visited.get(tuple(next_set), None):
                            step['next'] = next_set
                            for nsa in next_set:
                                self.steps[nsa]['needs'] = id
                            self.visited[tuple(curr_set)] = True

        # Finally, generate the root.
        for a_id, a_block in self.steps.items():
            if not a_block.get('needs', None):
                self.root.add(a_id)


VALID_WORKFLOW_ATTRS = ["resolves", "on"]


class HCLWorkflow(Workflow):
    """Parse a hcl based workflow and generate
    the workflow graph.
    """

    def __init__(self, wfile):
        """Loads the workflow as a dict from the `.workflow` file."""
        super(HCLWorkflow, self).__init__()
        self.wf_fmt = "hcl"

        if not os.path.exists(wfile):
            # try to load a string
            self.wf_dict = hcl.loads(wfile)
        else:
            with open(wfile) as fp:
                self.wf_dict = hcl.load(fp)

        if 'action' in self.wf_dict:
            self.wf_dict['steps'] = self.wf_dict.pop('action')

    def find_root(self, entrypoint, root):
        """A GHA workflow is defined by specifying edges that point to the
        previous nodes they depend on. To make the workflow easier to process,
        we add forward edges. This also obtains the root nodes.

        Args:
          entrypoint(list): List of nodes from where to start
        generating the graph.
          root(set): Set of nodes without dependencies,
        that would eventually be used as root.

        Returns:
            None
        """
        for node in entrypoint:
            if self.steps[node].get('needs', None):
                for n in self.steps[node]['needs']:
                    self.find_root([n], root)
                    if not self.steps[n].get('next', None):
                        self.steps[n]['next'] = set()
                    self.steps[n]['next'].add(node)
            else:
                root.add(node)

    def complete_graph(self):
        """Driver function to run the recursive function
        `find_root()` which adds forward edges.

        Args:
            None

        Returns:
            None
        """
        self.find_root(self.resolves, self.root)

    def normalize(self):
        """Takes properties from the `self.wf_dict` dict and makes them
        native to the `Workflow` class. Also it normalizes some of the
        attributes of a parsed workflow according to the Github defined
        specifications.

        For example, it changes `args`, `runs` and `secrets` attribute,
        if provided as a string to a list of string by splitting around
        whitespace. Also, it changes parameters like `uses` and `resolves`,
        if provided as a string to a list.

        Args:
            None

        Returns:
            None
        """
        for wf_name, wf_block in self.wf_dict['workflow'].items():

            self.name = wf_name
            self.resolves = wf_block['resolves']
            self.on = wf_block.get('on', 'push')
            self.root = set()
            self.steps = self.wf_dict['steps']
            self.props = dict()

            if pu.of_type(self.resolves, ['str']):
                self.resolves = [self.resolves]

        for a_name, a_block in self.steps.items():
            a_block['name'] = a_name

            if a_block.get('needs', None):
                if pu.of_type(a_block['needs'], ['str']):
                    a_block['needs'] = [a_block['needs']]

            if a_block.get('args', None):
                a_block['args'] = Workflow.format_command(a_block['args'])

            if a_block.get('runs', None):
                a_block['runs'] = Workflow.format_command(a_block['runs'])

            if a_block.get('secrets', None):
                a_block['secrets'] = Workflow.format_command(
                    a_block['secrets'])



import click
import os

from popper import log as logging
from popper.cli import log, pass_context
from popper.parser import Workflow
from popper.runner import WorkflowRunner


@click.command(
    'run', short_help='Run a workflow or step.')
@click.argument(
    'step',
    required=False
)
@click.option(
    '-f',
    '--wfile',
    help='File containing the definition of the workflow.',
    required=True
)
@click.option(
    '--debug',
    help=(
        'Generate detailed messages of what popper does (overrides --quiet)'),
    required=False,
    is_flag=True
)
@click.option(
    '--dry-run',
    help='Do not run the workflow, only print what would be executed.',
    required=False,
    is_flag=True
)
@click.option(
    '--log-file',
    help='Path to a log file. No log is created if this is not given.',
    required=False
)
@click.option(
    '--quiet',
    help='Do not print output generated by steps.',
    required=False,
    is_flag=True
)
@click.option(
    '--reuse',
    help='Reuse containers between executions (persist container state).',
    required=False,
    is_flag=True,
)
@click.option(
    '--engine',
    help='Specify runtime for executing the workflow [default: docker].',
    type=click.Choice(['docker', 'singularity', 'vagrant']),
    required=False,
    default='docker'
)
@click.option(
    '--skip',
    help=('Skip the given step (can be given multiple times).'),
    required=False,
    default=list(),
    hidden=True,
    multiple=True
)
@click.option(
    '--skip-clone',
    help='Skip cloning repositories (assume they have been cloned).',
    required=False,
    is_flag=True
)
@click.option(
    '--skip-pull',
    help='Skip pulling container images (assume they exist in local cache).',
    required=False,
    is_flag=True
)
@click.option(
    '--substitution',
    help=('A key-value pair defining a substitution. '
          'Can be given multiple times.'),
    required=False,
    default=list(),
    multiple=True
)
@click.option(
    '--allow-loose',
    help='Do not throw an error if a substitution variable passed as an '
    'argument is unused in the workflow definition.',
    required=False,
    is_flag=True
)
@click.option(
    '--with-dependencies',
    help='When STEP is given, execute all its dependencies as well.',
    required=False,
    hidden=True,
    is_flag=True
)
@click.option(
    '-w',
    '--workspace',
    help='Path to workspace folder.',
    required=False,
    show_default=False,
    default=os.getcwd()
)
@click.option(
    '-c', '--conf', help='Runtime configuration options.', required=False
)
@pass_context
def cli(ctx, step, wfile, debug, dry_run, log_file, quiet, reuse,
        engine, skip, skip_pull, skip_clone, substitution, allow_loose,
        with_dependencies, workspace, conf):
    """Runs a Popper workflow. Only execute STEP if given."""
    # set the logging levels.
    level = 'STEP_INFO'
    if quiet:
        level = 'INFO'
    if debug:
        level = 'DEBUG'
    log.setLevel(level)

    if dry_run:
        logging.msg_prefix = "DRYRUN: "

    if log_file:
        # also log to a file
        logging.add_log(log, log_file)

    # check conflicting flags and fail if needed
    if with_dependencies and not step:
        log.fail('`--with-dependencies` can only be used when '
                 'STEP argument is given.')
    if skip and step:
        log.fail('`--skip` can not be used when STEP argument is passed.')

    # invoke wf factory; handles formats, validations, filtering
    wf = Workflow.new(wfile, step=step, skipped_steps=skip,
                      substitutions=substitution, allow_loose=allow_loose,
                      include_step_dependencies=with_dependencies)

    # instantiate the runner
    runner = WorkflowRunner(config_file=conf, dry_run=dry_run, reuse=reuse,
                            skip_pull=skip_pull, skip_clone=skip_clone,
                            workspace_dir=workspace)
    runner.run(wf)



import os
import re

import git

from popper.cli import log


def new_repo():
    """Function to initialize a git.Repo object assumed to be in os.getcwd() or
    any parent directory.

    Args:
        None

    Returns:
        git.Repo: repo object or None if '.git/' not found in any parent folder
    """
    try:
        return git.Repo(search_parent_directories=True)
    except git.exc.InvalidGitRepositoryError:
        # Optimistically assume that this is due to .git/ folder not being
        # found, in which case all the methods in this module assume
        # os.getcwd() as the root of a project.
        pass
    return None


def get_project_root_folder(repo=None):
    """Function to find the project root folder.

    If the project is a git repository, the project root is the folder
    containing the `.git` folder, else the $PWD is assumed to be the
    project root.

    Args:
      None

    Returns:
      str: The path to the root of the project.

    """
    if repo:
        root_folder_path = repo.working_tree_dir
    else:
        root_folder_path = os.getcwd()

    return root_folder_path


def get_sha(repo):
    """Runs git rev-parse --short HEAD and returns result.

    This function returns 'unknown' if the project folder
    is not a git repo. It fails, when the project folder is a
    git repo but doesn't have any commit.

    Args:
        None

    Returns:
      str: The sha of the head commit or 'unknown'.

    """
    if repo:
        try:
            return repo.git.rev_parse(repo.head.object.hexsha, short=True)
        except ValueError as e:
            log.debug(e)
            log.fail(f'Could not obtain revision of repository located at {get_project_root_folder(repo)}'
                    )
    else:
        return 'na'


def get_remote_url(repo=None):
    """Obtains remote origin URL, if possible. Otherwise it returns empty str.

    Args:
        None

    Returns:
      str: The remote origin url or "".
    """
    url = ""
    if repo:
        if len(repo.remotes) > 0:
            url = repo.remotes.origin.url
        # cleanup the URL so we get in in https form and without '.git' ending
        if url.endswith('.git'):
            url = url[:-4]
        if 'git@' in url:
            url = 'https://' + url[4:].replace(':', '/')

    return url


def clone(url, org, repo, repo_dir, version=None):
    """Clones a repository using Git. The URL for the repo is
    https://github.com/ by default.

    Args:
      url(str): The remote git repository hosting service url.
      org(str): The org/user to which the repo belongs.
      repo(str): The repo name.
      repo_dir(str): The path where to clone the repo.
      version(str, optional): The remote tag/branch to checkout. If version is
                              None, we use the default remote branch as version

    Returns:
        None

    """
    def get_default_branch(r):
        """
        Used to find default branch of the repository.

        Args:
          r: It is git.Repo object

        Returns:

        """
        if version:
            return version
        return r.remotes.origin.refs['HEAD'].ref.remote_head

    if os.path.exists(repo_dir):
        cloned_repo = git.Repo(repo_dir)
        cloned_repo.remotes.origin.pull(get_default_branch(cloned_repo))
    else:
        if '@' in url:
            url += ':'
        else:
            url += '/'

        # To obtain the authentication token if set as environment variable.
        auth_token = os.getenv('GITHUB_API_TOKEN')

        if(auth_token is not None and 'github' in url and '@' not in url):
            # To verify the link of github for private repo support.
            # The authentication token has to be added after protocol
            # The length of protocol is 8 in case of https://
            url = url[:8]+auth_token+'@'+url[8:]

        repo_url = f'{url}{org}/{repo}'
        cloned_repo = git.Repo.clone_from(repo_url, repo_dir)

    cloned_repo.git.checkout(get_default_branch(cloned_repo))
    cloned_repo.close()


def parse(url):
    """Method to parse the git url.

    Args:
      url(str): The url in string format.

    Returns:
        tuple(service_url, service, user, repo, step_dir, version)

    """

    if url.startswith('ssh://'):
        log.fail('The ssh protocol is not supported yet.')

    if url.endswith('.git'):
        url = url[:-4]

    pattern = re.compile(
        r'^(http://|https://|git@)?(?:(\w+\.\w+)(?:\/|\:))?'
        r'([\w\-]+)(?:\/([^\@^\/]+)\/?([^\@]+)?(?:\@([\w\W]+))?)$')

    try:
        protocol, service, user, repo, step_dir, version = pattern.search(
            url).groups()
    except AttributeError:
        log.fail(
            'Invalid url. The url should be in any of the 3 forms: \n'
            '1) https://github.com/user/repo/path/to/step@version \n'
            '2) gitlab.com/user/repo/path/to/step@version \n'
            '3) user/repo/path/to/step@version'
        )

    if not service:
        service = 'github.com'

    if not protocol:
        protocol = 'https://'

    if not step_dir:
        step_dir = ''

    service_url = protocol + service

    log.debug(f'parse("{url}"):')
    log.debug(f'  service_url: {service_url}')
    log.debug(f'  service: {service}')
    log.debug(f'  user: {user}')
    log.debug(f'  repo: {repo}')
    log.debug(f'  step_dir: {step_dir}')
    log.debug(f'  version: {version}')

    return service_url, service, user, repo, step_dir, version



import os

import click

from popper.cli import pass_context, log


@click.command('scaffold', short_help='Generate a minimal workflow.')
@click.option(
    '-f',
    '--wfile',
    help='Name of file where to write the generated workflow.',
    required=False,
    default='wf.yml'
)
@pass_context
def cli(ctx, wfile):
    """Generates a minimal workflow that can be used as starting point."""
    main_workflow_content = """steps:

- uses: "popperized/bin/sh@master"
  args: "ls"

- uses: "docker://alpine:3.11"
  args: ["ls"]
"""

    if os.path.exists(wfile):
        log.fail(f'File {wfile} already exists')

    with open(wfile, 'w') as f:
        f.write(main_workflow_content)

    log.info('Successfully generated a workflow scaffold.')



import os

from subprocess import PIPE, Popen, STDOUT, SubprocessError

from popper import utils as pu
from popper.cli import log as log
from popper.runner import StepRunner as StepRunner



class HostRunner(StepRunner):
    """Run an step on the Host Machine."""

    spawned_processes = []

    def __init__(self, config):
        super(HostRunner, self).__init__(config)
        if self.config.reuse:
            log.warning('Reuse not supported for HostRunner.')

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, traceback):
        pass

    def run(self, step):
        step_env = StepRunner.prepare_environment(step, os.environ)

        cmd = step.get('runs', [])
        if not cmd:
            raise AttributeError(f"Expecting 'runs' attribute in step.")
        cmd.extend(step.get('args', []))

        log.info(f'[{step["name"]}] {cmd}')

        if self.config.dry_run:
            return 0

        log.debug(f'Environment:\n{pu.prettystr(os.environ)}')

        try:
            with Popen(cmd, stdout=PIPE, stderr=STDOUT,
                       universal_newlines=True, preexec_fn=os.setsid,
                       env=step_env, cwd=self.config.workspace_dir) as p:
                HostRunner.spawned_processes.append(p)

                log.debug('Reading process output')

                for line in iter(p.stdout.readline, ''):
                    line_decoded = pu.decode(line)
                    log.step_info(line_decoded[:-1])

                p.wait()
                ecode = p.poll()

            log.debug(f'Code returned by process: {ecode}')

        except SubprocessError as ex:
            ecode = ex.returncode
            log.step_info(f"Command '{cmd[0]}' failed with: {ex}")
        except Exception as ex:
            ecode = 1
            log.step_info(f"Command raised non-SubprocessError error: {ex}")

        return ecode

    def stop_running_tasks(self):
        for p in HostRunner.spawned_processes:
            log.info(f'Stopping proces {p.pid}')
            p.kill()



import os
import shutil
import tempfile
import unittest

import git

from popper import scm
from popper.cli import log


class TestScm(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        log.setLevel('CRITICAL')

        self.tempdir = tempfile.mkdtemp()
        self.curr_dir = os.getcwd()

        self.repo = git.Repo.clone_from('https://github.com/popperized/bin',
                                        os.path.join(self.tempdir, 'bin'))

        self.repodir = os.path.join(self.tempdir, 'bin')

        os.chdir(self.repodir)

        self.gitdir = os.path.join(self.repodir, '.git')
        self.gotdor = os.path.join(self.repodir, '.got')

    @classmethod
    def tearDownClass(self):
        log.setLevel('NOTSET')

        # return to where we were before this test
        os.chdir(self.curr_dir)

        self.repo.close()

    def test_with_git(self):

        if not os.path.exists(self.gitdir):
            shutil.move(self.gotdor, self.gitdir)

        # root folder
        root_folder = scm.get_project_root_folder(self.repo)
        self.assertEqual(os.path.realpath(root_folder),
                         os.path.realpath(os.path.join(self.tempdir, 'bin')))

        # get_remote_url
        url = scm.get_remote_url(self.repo)
        auth_token = os.getenv('GITHUB_API_TOKEN')
        if not auth_token:
            self.assertEqual(url, 'https://github.com/popperized/bin')
        else:
            self.assertTrue('github.com/popperized/bin' in url)

        # get sha
        sha = scm.get_sha(self.repo)
        expected = self.repo.git.rev_parse(self.repo.head.object.hexsha,
                                           short=True)
        self.assertEqual(sha, expected)

    def test_without_git(self):
        shutil.move(self.gitdir, self.gotdor)

        # root folder
        root_folder = scm.get_project_root_folder(None)
        self.assertEqual(os.path.realpath(root_folder),
                         os.path.realpath(os.path.join(self.tempdir, 'bin')))

        # get_remote_url
        self.assertEqual(scm.get_remote_url(None), '')

        # get sha
        sha = scm.get_sha(None)
        self.assertEqual(sha, 'na')

    def test_clone(self):
        tdir = os.path.join(self.tempdir, 'test_clone')
        os.makedirs(tdir)
        currdir = os.getcwd()
        os.chdir(tdir)
        scm.clone(
            'https://github.com',
            'popperized',
            'github-actions-demo',
            os.path.join(os.getcwd(), 'gad'),
            'develop'
        )
        repo = git.Repo(os.path.join(os.getcwd(), 'gad'))
        self.assertEqual(repo.active_branch.name, 'develop')
        scm.clone(
            'https://github.com',
            'popperized',
            'github-actions-demo',
            os.path.join(os.getcwd(), 'gad'),
            'master'
        )
        repo = git.Repo(os.path.join(os.getcwd(), 'gad'))
        self.assertEqual(repo.active_branch.name, 'master')
        repo.close()
        os.chdir(currdir)

    def test_parse(self):
        test_url = ("ssh://git@github.com:popperized"
                    "/github-actions-demo.git")
        self.assertRaises(SystemExit, scm.parse, test_url)
        test_url = "https://github.com/popperized@master"
        self.assertRaises(SystemExit, scm.parse, test_url)
        test_url = "github.com"
        self.assertRaises(SystemExit, scm.parse, test_url)
        test_url = "@master"
        self.assertRaises(SystemExit, scm.parse, test_url)
        test_url = "http://gitlab.com/popperized/github-actions-demo.git"
        parts = scm.parse(test_url)
        self.assertTupleEqual(parts, (
            'http://gitlab.com',
            'gitlab.com',
            'popperized',
            'github-actions-demo',
            '', None))
        test_url = ("https://github.com/popperized"
                    "/github-actions-demo@master")
        parts = scm.parse(test_url)
        self.assertTupleEqual(parts, (
            'https://github.com',
            'github.com',
            'popperized',
            'github-actions-demo',
            '',
            'master'))
        test_url = "github.com/popperized/github-actions-demo"
        parts = scm.parse(test_url)
        self.assertTupleEqual(parts, (
            'https://github.com',
            'github.com',
            'popperized',
            'github-actions-demo',
            '',
            None))
        test_url = "popperized/github-actions-demo/path/to/action@develop"
        parts = scm.parse(test_url)
        self.assertTupleEqual(parts, (
            'https://github.com',
            'github.com',
            'popperized',
            'github-actions-demo',
            'path/to/action',
            'develop'))



import unittest
import os

from popper import utils as pu
from popper.cli import log


class TestUtils(unittest.TestCase):

    def setUp(self):
        log.setLevel('CRITICAL')

    def tearDown(self):
        log.setLevel('NOTSET')

    def test_decode(self):
        string = b'Hello from popper'
        result = pu.decode(string)
        self.assertIsInstance(result, str)

        string = 'Hello from popper'
        result = pu.decode(string)
        self.assertIsInstance(result, str)

    def test_sanitized_name(self):
        name = "test action"
        santizied_name = pu.sanitized_name(name, '1234')
        self.assertEqual(santizied_name, "popper_test_action_1234")

        name = "test@action"
        santizied_name = pu.sanitized_name(name, '1234')
        self.assertEqual(santizied_name, "popper_test_action_1234")

        name = "test(action)"
        santizied_name = pu.sanitized_name(name, '1234')
        self.assertEqual(santizied_name, "popper_test_action__1234")

    def touch_file(self, path):
        open(path, 'w').close()

    def test_setup_base_cache(self):
        cache_dir = pu.setup_base_cache()
        try:
            self.assertEqual(cache_dir, os.environ['XDG_CACHE_HOME'])
        except KeyError:
            self.assertEqual(
                cache_dir,
                os.path.join(
                    os.environ['HOME'],
                    '.cache/popper'))

        os.environ['POPPER_CACHE_DIR'] = '/tmp/popper'
        cache_dir = pu.setup_base_cache()
        self.assertEqual(cache_dir, '/tmp/popper')
        os.environ.pop('POPPER_CACHE_DIR')

    def test_of_type(self):
        param = [u"hello", u"world"]
        self.assertEqual(pu.of_type(param, ['los']), True)

        param = u"hello world"
        self.assertEqual(pu.of_type(param, ['los']), False)

        param = {
            "org": "systemslab",
            "project": "popper"
        }
        self.assertEqual(pu.of_type(param, ['str', 'dict']), True)

    def test_write_file(self):
        content = "Hello world"
        pu.write_file('testfile1.txt', content)
        pu.write_file('testfile2.txt')
        with open('testfile1.txt', 'r') as f:
            self.assertEqual(f.read(), "Hello world")
        with open('testfile2.txt', 'r') as f:
            self.assertEqual(f.read(), '')
        os.remove('testfile1.txt')
        os.remove('testfile2.txt')

    def test_load_config_file(self):
        conf_content = """ENGINE = {
    "runtime": "nvidia"
}
        """
        pu.write_file('settings.py', conf_content)
        config = pu.load_config_file('settings.py')
        self.assertTrue(hasattr(config, 'ENGINE'))
        self.assertDictEqual(config.ENGINE, {'runtime': 'nvidia'})
        os.remove('settings.py')

    def test_assert_executable_exists(self):
        self.assertRaises(SystemExit, pu.assert_executable_exists, 'abcd')



import os
import unittest

import utils as testutils

from popper.parser import YMLWorkflow
from popper.runner import WorkflowRunner

from popper.cli import log as log


class TestHostRunner(unittest.TestCase):
    def setUp(self):
        log.setLevel('CRITICAL')

    def test_run(self):
        repo = testutils.mk_repo()

        with WorkflowRunner(workspace_dir=repo.working_dir) as r:
            wf = YMLWorkflow("""
            version: '1'
            steps:
            - uses: sh
              runs: [cat]
              args: README.md
            """)
            wf.parse()
            r.run(wf)

            wf = YMLWorkflow("""
            version: '1'
            steps:
            - uses: 'sh'
              runs: ['bash', '-c', 'echo $FOO > hello.txt ; pwd']
              env: {
                  FOO: bar
              }
            """)
            wf.parse()
            r.run(wf)
            with open(os.path.join(repo.working_dir, 'hello.txt'), 'r') as f:
                self.assertEqual(f.read(), 'bar\n')

            wf = YMLWorkflow("""
            version: '1'
            steps:
            - uses: 'sh'
              runs: 'nocommandisnamedlikethis'
            """)
            wf.parse()
            self.assertRaises(SystemExit, r.run, wf)

        repo.close()



import unittest
import os

from popper.parser import Workflow, YMLWorkflow, HCLWorkflow
from popper.cli import log

FIXDIR = f'{os.path.dirname(os.path.realpath(__file__))}/fixtures'


def _wfile(name, format):
    return f'{FIXDIR}/{name}.{format}'


class TestWorkflow(unittest.TestCase):
    def setUp(self):
        log.setLevel('CRITICAL')

    def tearDown(self):
        log.setLevel('NOTSET')

    def test_new_workflow(self):
        self.assertIsInstance(
            Workflow.new(_wfile('a', 'yml')), YMLWorkflow)
        self.assertIsInstance(
            Workflow.new(_wfile('a', 'workflow')), HCLWorkflow)

    def test_missing_dependency(self):
        wf = HCLWorkflow(_wfile('missing_dependency', 'workflow'))
        wf.normalize()
        self.assertRaises(SystemExit, wf.check_for_broken_workflow)
        wf = YMLWorkflow(_wfile('missing_dependency', 'yml'))
        wf.normalize()
        self.assertRaises(SystemExit, wf.check_for_broken_workflow)

    def test_command(self):
        cmd = u"docker version"
        res = Workflow.format_command(cmd)
        self.assertEqual(res, ["docker", "version"])

        cmd = ["docker", "version"]
        res = Workflow.format_command(cmd)
        self.assertEqual(res, ["docker", "version"])

    def test_validate_workflow_block(self):
        wf = HCLWorkflow("""workflow "w1" {
    resolves = ["a"]
}
workflow "w2" {
    resolves = ["a"]
}
""")
        self.assertRaises(SystemExit, wf.validate_workflow_block)

        wf = HCLWorkflow("""
action "a" {
    uses = "sh"
}
""")
        self.assertRaises(SystemExit, wf.validate_workflow_block)

        wf = HCLWorkflow("""
workflow "sample workflow 1" {
    resolves = ["a"]
    runs = ["sh", "-c", "ls"]
}
action "a" {
    uses = ["sh"]
}
""")
        self.assertRaises(SystemExit, wf.validate_workflow_block)

        wf = HCLWorkflow("""
workflow "sample workflow 1" {
    on = "push"
}
action "a" {
    uses = ["sh"]
}
""")
        self.assertRaises(SystemExit, wf.validate_workflow_block)

    def test_validate_step_blocks(self):
        wf = HCLWorkflow("""workflow "sample workflow" {
            resolves = "a"
        }""")
        self.assertRaises(SystemExit, wf.validate_step_blocks)

        wf = HCLWorkflow("""workflow "sample workflow" {
    resolves = "a"
}
action "a" {
    uses = "sh"
    on = "push"
}""")
        self.assertRaises(SystemExit, wf.validate_step_blocks)

        wf = HCLWorkflow("""workflow "sample workflow" {
    resolves = "a"
}
action "a" {
    args = "ls"
}""")
        self.assertRaises(SystemExit, wf.validate_step_blocks)

        wf = HCLWorkflow("""workflow "sample workflow" {
    resolves = "a"
}
action "a" {
    uses = 1
}""")
        self.assertRaises(SystemExit, wf.validate_step_blocks)

        wf = HCLWorkflow("""workflow "sample workflow" {
    resolves = "a"
}

action "a" {
    uses = "sh"
    needs = 1
}""")
        self.assertRaises(SystemExit, wf.validate_step_blocks)

        wf = HCLWorkflow("""workflow "sample workflow" {
    resolves = "a"
}
action "a" {
    uses = "sh"
    args = [1, 2, 3, 4]
}
""")
        self.assertRaises(SystemExit, wf.validate_step_blocks)

        wf = HCLWorkflow("""workflow "sample workflow" {
    resolves = "a"
}
action "a" {
    uses = "sh"
    runs = [1, 2, 3, 4]
}
""")
        self.assertRaises(SystemExit, wf.validate_step_blocks)

        wf = HCLWorkflow("""workflow "sample workflow" {
    resolves = "a"
}
action "a" {
    uses = "sh"
    secrets = {
        SECRET_A = 1234,
        SECRET_B =  5678
    }
}
""")
        self.assertRaises(SystemExit, wf.validate_step_blocks)

        wf = HCLWorkflow("""workflow "sample workflow" {
    resolves = "a"
}
action "a" {
    uses = "sh"
    env = [
        "SECRET_A", "SECRET_B"
    ]
}
""")
        self.assertRaises(SystemExit, wf.validate_step_blocks)

    def test_skip_steps(self):
        wf = YMLWorkflow(_wfile('a', 'yml'))
        wf.parse()
        changed_wf = Workflow.skip_steps(wf, ['b'])
        self.assertDictEqual(changed_wf.steps, {
            'a': {
                'uses': 'sh',
                'args': ['ls'],
                'name': 'a',
                'next': {'e'}},
            'b': {
                'uses': 'sh',
                'args': ['ls'],
                'name': 'b',
                'next': set()},
            'c': {
                'uses': 'sh',
                'args': ['ls'],
                'name': 'c',
                'next': {'d'}},
            'd': {
                'needs': ['c'],
                'uses': 'sh',
                'args': ['ls'],
                'name': 'd',
                'next': {'e'}},
            'e': {
                'needs': ['d', 'a'],
                'uses': 'sh',
                'args': ['ls'],
                'name': 'e',
                'next': {'end'}},
            'end': {
                'needs': ['e'],
                'uses': 'sh',
                'args': ['ls'],
                'name': 'end'}})

        changed_wf = Workflow.skip_steps(wf, ['d', 'a'])
        self.assertDictEqual(changed_wf.steps, {
            'a': {
                'uses': 'sh',
                'args': ['ls'],
                'name': 'a',
                'next': set()},
            'b': {
                'uses': 'sh',
                'args': ['ls'],
                'name': 'b',
                'next': {'e'}},
            'c': {
                'uses': 'sh',
                'args': ['ls'],
                'name': 'c',
                'next': set()},
            'd': {
                'needs': [],
                'uses': 'sh',
                'args': ['ls'],
                'name': 'd',
                'next': set()},
            'e': {
                'needs': ['b'],
                'uses': 'sh',
                'args': ['ls'],
                'name': 'e',
                'next': {'end'}},
            'end': {
                'needs': ['e'],
                'uses': 'sh',
                'args': ['ls'],
                'name': 'end'}})

    def test_filter_step(self):
        wf = YMLWorkflow(_wfile('a', 'yml'))
        wf.parse()
        changed_wf = Workflow.filter_step(wf, 'e')
        self.assertSetEqual(changed_wf.root, {'e'})
        self.assertDictEqual(
            changed_wf.steps, {
                'e': {
                    'needs': [],
                    'uses': 'sh',
                    'args': ['ls'],
                    'name': 'e',
                    'next': set()}})

        changed_wf = Workflow.filter_step(wf, 'd')
        self.assertSetEqual(changed_wf.root, {'d'})
        self.assertDictEqual(
            changed_wf.steps, {
                'd': {
                    'needs': [],
                    'uses': 'sh',
                    'args': ['ls'],
                    'name': 'd',
                    'next': set()}})

        changed_wf = Workflow.filter_step(wf, 'e', with_dependencies=True)
        self.assertSetEqual(changed_wf.root, {'b', 'a', 'c'})
        self.assertDictEqual(changed_wf.steps, {
            'a': {
                'uses': 'sh',
                'args': ['ls'],
                'name': 'a',
                'next': {'e'}},
            'b': {
                'uses': 'sh',
                'args': ['ls'],
                'name': 'b',
                'next': {'e'}},
            'c': {
                'uses': 'sh',
                'args': ['ls'],
                'name': 'c',
                'next': {'d'}},
            'd': {
                'needs': ['c'],
                'uses': 'sh',
                'args': ['ls'],
                'name': 'd',
                'next': {'e'}},
            'e': {
                'needs': ['d', 'b', 'a'],
                'uses': 'sh',
                'args': ['ls'],
                'name': 'e',
                'next': set()}})

        changed_wf = Workflow.filter_step(wf, 'd', with_dependencies=True)
        self.assertSetEqual(changed_wf.root, {'c'})
        self.assertDictEqual(
            changed_wf.steps, {
                'c': {
                    'uses': 'sh',
                    'args': ['ls'],
                    'name': 'c',
                    'next': {'d'}},
                'd': {
                    'needs': ['c'],
                    'uses': 'sh',
                    'args': ['ls'],
                    'name': 'd',
                    'next': set()}})

    def test_check_for_unreachable_steps(self):
        wf = HCLWorkflow(_wfile('a', 'workflow'))
        wf.parse()
        changed_wf = Workflow.skip_steps(wf, ['d', 'a', 'b'])
        self.assertDictEqual(changed_wf.steps, {
            'a': {
                'uses': 'sh',
                'args': ['ls'],
                'name': 'a',
                'next': set()},
            'b': {
                'uses': 'sh',
                'args': ['ls'],
                'name': 'b',
                'next': set()},
            'c': {
                'uses': 'sh',
                'args': ['ls'],
                'name': 'c',
                'next': set()},
            'd': {
                'needs': [],
                'uses': 'sh',
                'args': ['ls'],
                'name': 'd',
                'next': set()},
            'e': {
                'needs': [],
                'uses': 'sh',
                'args': ['ls'],
                'name': 'e',
                'next': {'end'}},
            'end': {
                'needs': ['e'],
                'uses': 'sh',
                'args': ['ls'],
                'name': 'end'}
        })

        changed_wf.check_for_unreachable_steps()

        wf = HCLWorkflow(_wfile('ok', 'workflow'))
        wf.parse()
        wf.check_for_unreachable_steps()

    def test_get_stages(self):
        wf = HCLWorkflow(_wfile('a', 'workflow'))
        wf.parse()
        stages = list()
        for stage in wf.get_stages():
            stages.append(stage)

        self.assertListEqual(stages, [
            {'b', 'c', 'a'},
            {'d'},
            {'e'},
            {'end'}
        ])

        wf = YMLWorkflow(_wfile('b', 'yml'))
        wf.parse()
        stages = list()
        for stage in wf.get_stages():
            stages.append(stage)

        self.assertListEqual(stages, [
            {'a', 'c'},
            {'b'},
            {'d'},
            {'g', 'f'},
            {'h'},
            {'end'}
        ])

    def test_substitutions(self):
        subs = [
            '_VAR1=sh', '_VAR2=ls', '_VAR4=test_env',
            '_VAR5=TESTING', '_VAR6=TESTER', '_VAR7=TEST'
        ]
        wf = YMLWorkflow(_wfile('substitutions', 'yml'))
        wf.parse(subs, False)
        self.assertDictEqual(wf.steps, {
            'a': {
                'uses': 'sh',
                'args': ['ls'],
                'name': 'a',
                'next': {'b'}},
            'b': {
                'needs': ['a'],
                'uses': 'sh',
                'args': ['ls'],
                'runs': ['test_env'],
                'secrets': ['TESTING'],
                'env': {
                    'TESTER': 'TEST'
                },
                'name': 'b'}
        })

        wf = YMLWorkflow(_wfile('substitutions', 'yml'))
        wf.parse(subs, False)
        self.assertDictEqual(wf.steps, {
            'a': {
                'uses': 'sh',
                'args': ['ls'],
                'name': 'a',
                'next': {'b'}},
            'b': {
                'needs': ['a'],
                'uses': 'sh',
                'args': ['ls'],
                'runs': ['test_env'],
                'secrets': ['TESTING'],
                'env': {
                    'TESTER': 'TEST'
                },
                'name': 'b'}
        })


class TestHCLWorkflow(unittest.TestCase):
    def setUp(self):
        log.setLevel('CRITICAL')

    def tearDown(self):
        log.setLevel('NOTSET')

    def test_load_file(self):
        wf = HCLWorkflow("""workflow "sample" {
    resolves = "b"
}
action "a" {
    uses = "sh"
}
action "b" {
    needs = "a"
    uses = "sh"
}""")
        self.assertEqual(wf.wf_fmt, "hcl")
        self.assertDictEqual(
            wf.wf_dict, {
                'workflow': {
                    'sample': {
                        'resolves': 'b'}}, 'steps': {
                    'a': {
                        'uses': 'sh'}, 'b': {
                            'needs': 'a', 'uses': 'sh'}}})

    def test_normalize(self):
        wf = HCLWorkflow("""workflow "sample workflow" {
    resolves = "a"
}
action "a" {
    needs = "b"
    uses = "popperized/bin/npm@master"
    args = "npm --version"
    secrets = "SECRET_KEY"
}""")
        wf.normalize()
        self.assertEqual(wf.resolves, ['a'])
        self.assertEqual(wf.name, 'sample workflow')
        self.assertEqual(wf.on, 'push')
        self.assertDictEqual(wf.props, dict())
        step_a = wf.steps['a']
        self.assertEqual(step_a['name'], 'a')
        self.assertEqual(step_a['needs'], ['b'])
        self.assertEqual(step_a['args'], ['npm', '--version'])
        self.assertEqual(step_a['secrets'], ['SECRET_KEY'])

    def test_complete_graph(self):
        wf = HCLWorkflow(_wfile('a', 'workflow'))
        wf.normalize()
        wf.complete_graph()
        self.assertEqual(wf.name, 'example')
        self.assertEqual(wf.resolves, ['end'])
        self.assertEqual(wf.on, 'push')
        self.assertEqual(wf.props, {})
        self.assertEqual(wf.root, {'b', 'c', 'a'})

        steps_dict = {
            'a': {
                'uses': 'sh',
                'args': ['ls'],
                'name': 'a',
                'next': {'e'}},
            'b': {
                'uses': 'sh',
                'args': ['ls'],
                'name': 'b',
                'next': {'e'}},
            'c': {
                'uses': 'sh',
                'args': ['ls'],
                'name': 'c',
                'next': {'d'}},
            'd': {
                'needs': ['c'],
                'uses': 'sh',
                'args': ['ls'],
                'name': 'd',
                'next': {'e'}},
            'e': {
                'needs': ['d', 'b', 'a'],
                'uses': 'sh',
                'args': ['ls'],
                'name': 'e',
                'next': {'end'}},
            'end': {
                'needs': ['e'],
                'uses': 'sh',
                'args': ['ls'],
                'name': 'end'}
        }
        self.assertDictEqual(wf.steps, steps_dict)


class TestYMLWorkflow(unittest.TestCase):
    def setUp(self):
        log.setLevel('CRITICAL')

    def tearDown(self):
        log.setLevel('NOTSET')

    def test_load_file(self):
        wf = YMLWorkflow("""
        steps:
        - id: 'a'
          uses: 'sh'

        - id: 'b'
          uses: 'sh'
        """)
        self.assertEqual(wf.wf_fmt, "yml")
        self.assertDictEqual(
            wf.wf_dict, {
                'steps': {
                    'a': {
                        'uses': 'sh'}, 'b': {
                        'uses': 'sh'}}})
        self.assertListEqual(
            wf.wf_list,
            [{'uses': 'sh'}, {'uses': 'sh'}])
        self.assertDictEqual(
            wf.id_map,
            {1: 'a', 2: 'b'})

    def test_normalize(self):
        wf = YMLWorkflow("""
        steps:
        - id: "a"
          needs: "b"
          uses: "popperized/bin/npm@master"
          args: "npm --version"
          secrets: "SECRET_KEY"
        """)
        wf.normalize()
        self.assertEqual(wf.on, '')
        self.assertDictEqual(wf.props, dict())
        step_a = wf.steps['a']
        self.assertEqual(step_a['name'], 'a')
        self.assertEqual(step_a['needs'], ['b'])
        self.assertEqual(step_a['args'], ['npm', '--version'])
        self.assertEqual(step_a['secrets'], ['SECRET_KEY'])

    def test_get_containing_set(self):
        wf = YMLWorkflow(_wfile('a', 'yml'))
        wf.normalize()
        wf.complete_graph()
        set_1 = wf.get_containing_set(2)
        self.assertSetEqual(set_1, {'b', 'a', 'd'})

        wf = YMLWorkflow(_wfile('b', 'yml'))
        wf.normalize()
        wf.complete_graph()
        set_2 = wf.get_containing_set(3)
        self.assertSetEqual(set_2, {'c', 'b'})

    def test_complete_graph(self):
        wf = YMLWorkflow(_wfile('a', 'yml'))
        wf.normalize()
        wf.complete_graph()
        self.assertEqual(wf.name, 'a')
        self.assertEqual(wf.on, '')
        self.assertEqual(wf.props, {})
        self.assertEqual(wf.root, {'b', 'c', 'a'})

        steps_dict = {
            'a': {
                'uses': 'sh',
                'args': ['ls'],
                'name': 'a',
                'next': {'e'}
            },
            'b': {
                'uses': 'sh',
                'args': ['ls'],
                'name': 'b',
                'next': {'e'}
            },
            'c': {
                'uses': 'sh',
                'args': ['ls'],
                'name': 'c',
                'next': {'d'}
            }, 'd': {
                'needs': ['c'],
                'uses': 'sh',
                'args': ['ls'],
                'name': 'd',
                'next': {'e'}
            }, 'e': {
                'needs': ['d', 'b', 'a'],
                'uses': 'sh',
                'args': ['ls'],
                'name': 'e',
                'next': {'end'}
            }, 'end': {
                'needs': ['e'],
                'uses': 'sh',
                'args': ['ls'],
                'name': 'end'}
        }
        self.assertDictEqual(wf.steps, steps_dict)



import click
import os

from popper.cli import pass_context, log

ci_files = {

    # ##########################################################3

    'travis': {
        './.travis.yml': """
---
dist: xenial
language: python
python: 3.7
services: docker
install:
- git clone https://github.com/systemslab/popper /tmp/popper
- export PYTHONUNBUFFERED=1
- pip install /tmp/popper/cli
script: popper run -f {}
"""
    },

    # ##########################################################3

    'circle': {
        './.circleci/config.yml': """
---
version: 2
jobs:
  build:
    machine: true
    steps:
    - checkout
    - run:
        command: |
        git clone https://github.com/systemslab/popper /tmp/popper
        export PYTHONUNBUFFERED=1
        pip install /tmp/popper/cli
        popper run -f {}
"""
    },

    # ##########################################################3

    'jenkins': {
        './Jenkinsfile': """
---
stage ('Popper') {{ node {{
  sh "git clone https://github.com/systemslab/popper /tmp/popper"
  sh "export PYTHONUNBUFFERED=1"
  sh "pip install /tmp/popper/cli"
  sh "popper run -f {}"
}}}}
"""
    },

    # ##########################################################3

    'gitlab': {
        '.gitlab-ci.yml': """
---
image: docker:stable

variables:
  DOCKER_DRIVER: overlay

services:
- docker:dind

before_script:
- docker info
- apk update
- apk upgrade
- apk add python python-dev py-pip build-base git bash
- pip install virtualenv
- git clone https://github.com/systemslab/popper /tmp/popper
- pip install /tmp/popper/cli
popper:
  script: popper run -f {}
"""
    },

    # ##########################################################3

    'brigade': {
        'brigade.js': """
const { events, Job } = require("brigadier")
events.on("push", () => {
    var popper = new Job("popper", "python:3.7-slim-stretch")
    popper.tasks = [
        "apt-get update",
        "apt-get install -y git",
        "git clone https://github.com/systemslab/popper /tmp/popper",
        "export PYTHONUNBUFFERED=1",
        "pip install /tmp/popper/cli",
        "popper run -f {}"
    ]
    popper.run()
})
"""
    }
}


@click.command('ci', short_help='Generate CI service configuration files.')
@click.argument(
    'service',
    type=click.Choice(['travis', 'circle', 'jenkins', 'gitlab', 'brigade']),
    required=True
)
@click.option(
    '-f', '--wfile', help='Specify workflow to run in CI.', required=True
)
@pass_context
def cli(ctx, service, wfile):
    """Generates configuration files for distinct CI services. This command
    needs to be executed on the root of your Git repository folder.
    """
    if not os.path.exists('.git'):
        log.fail(
            'This command needs to be executed on the root of your '
            'Git project folder (where the .git/ folder is located).')

    for ci_file, ci_file_content in ci_files[service].items():
        ci_file = os.path.join(os.getcwd(), ci_file)
        os.makedirs(os.path.dirname(ci_file), exist_ok=True)
        with open(ci_file, 'w') as f:
            f.write(ci_file_content.format(wfile))

    log.info(f'Wrote {service} configuration successfully.')



import os
import unittest
import shutil

from dotmap import DotMap
from unittest.mock import patch
from popper.parser import YMLWorkflow
from popper.runner import WorkflowRunner, StepRunner

from popper.cli import log


class TestWorkflowRunner(unittest.TestCase):
    def setUp(self):
        log.setLevel('CRITICAL')

    def test_check_secrets(self):
        wf = YMLWorkflow("""
        version: '1'
        steps:
        - uses: docker://alpine:3.9
          args: ["ls -ltr"]
          secrets: ["SECRET_ONE", "SECRET_TWO"]
        """)
        wf.parse()

        # in dry-run, secrets are ignored
        WorkflowRunner.process_secrets(wf, DotMap({'dry_run': True}))

        # when CI=true it should fail
        os.environ['CI'] = 'true'
        self.assertRaises(SystemExit, WorkflowRunner.process_secrets, wf)

        # add one secret
        os.environ['SECRET_ONE'] = '1234'

        # it should fail again, as we're missing one
        self.assertRaises(SystemExit, WorkflowRunner.process_secrets, wf)

        os.environ.pop('CI')

        # now is fine
        with patch('getpass.getpass', return_value='5678'):
            WorkflowRunner.process_secrets(wf)

        # pop the other
        os.environ.pop('SECRET_ONE')

    def test_clone_repos(self):
        wf = YMLWorkflow("""
        version: '1'
        steps:
        - uses: popperized/bin/sh@master
        """)
        wf.parse()

        wid = '12345'
        cache_dir = os.path.join(os.environ['HOME'], '.cache/popper/')

        # clone repos in the default cache directory.
        WorkflowRunner.clone_repos(wf, DotMap({'wid': wid}))
        self.assertTrue(
            os.path.exists(
                os.path.join(cache_dir, wid, 'github.com/popperized/bin')))

        # clone repos in custom cache directory
        os.environ['POPPER_CACHE_DIR'] = '/tmp/smdir'
        WorkflowRunner.clone_repos(wf, DotMap({'wid': wid}))
        self.assertTrue(
            os.path.exists(
                os.path.join('/tmp/smdir', wid, 'github.com/popperized/bin')))
        os.environ.pop('POPPER_CACHE_DIR')

        # check failure when container is not available and we skip cloning
        shutil.rmtree('/tmp/smdir')
        shutil.rmtree(cache_dir)
        self.assertRaises(SystemExit, WorkflowRunner.clone_repos, wf,
                          DotMap({'wid': wid, 'skip_clone': True}))

    def test_steprunner_factory(self):
        with WorkflowRunner() as r:
            self.assertEqual(r.get_step_runner('host').__class__.__name__,
                             'HostRunner')
            self.assertEqual(r.get_step_runner('docker').__class__.__name__,
                             'DockerRunner')


class TestStepRunner(unittest.TestCase):
    def setUp(self):
        log.setLevel('CRITICAL')

    def test_prepare_environment(self):
        step = {'name': 'a', 'env': {'FOO': 'BAR'}, 'secrets': ['A']}
        os.environ['A'] = 'BC'
        env = StepRunner.prepare_environment(step, {'another': 'b'})
        self.assertDictEqual(env, {'FOO': 'BAR', 'A': 'BC', 'another': 'b'})
        os.environ.pop('A')



import getpass
import os
import sys

from dotmap import DotMap
from hashlib import shake_256

import popper.scm as scm
import popper.utils as pu

from popper.cli import log


class WorkflowRunner(object):
    """The workflow runner."""

    # class variable that holds references to runner singletons
    runners = {}

    def __init__(self, config_file=None, workspace_dir=os.getcwd(),
                 reuse=False, engine_name='docker', dry_run=False, quiet=False,
                 skip_pull=False, skip_clone=False):

        # save all args in a member dictionary
        self.config = DotMap(locals())
        self.config.pop('self')
        self.config.workspace_dir = os.path.realpath(workspace_dir)
        self.config.wid = shake_256(workspace_dir.encode('utf-8')).hexdigest(4)

        # cretate a repo object for the project
        self.repo = scm.new_repo()
        self.config.workspace_sha = scm.get_sha(self.repo)

        if config_file:
            # read options from config file
            config_from_file = pu.load_config_file(config_file)

            if hasattr(config_from_file, 'ENGINE'):
                self.config.engine_options = config_from_file.ENGINE

        import popper.runner_docker
        import popper.runner_host
        # import popper.runner_singularity
        # import popper.runner_vagrant

        # create dict with runner classes
        self.runner_classes = {
            'docker': popper.runner_docker.DockerRunner,
            'host': popper.runner_host.HostRunner,
            # 'singularity': popper.runner_singularity.SingularityRunner,
            # 'vagrant': popper.runner_vagrant.VagrantRunner,
        }

        # check that given engine is supported
        if engine_name not in self.runner_classes:
            raise ValueError(f"Invalid value for engine: '{engine_name}'.")

        # holds references to (singleton) instances of runners
        WorkflowRunner.runners = {}

        log.debug(f'WorkflowRunner config:\n{pu.prettystr(self.config)}')

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, traceback):
        """calls __exit__ on all instantiated step runners"""
        self.repo.close()
        for _, r in self.runners.items():
            r.__exit__(exc_type, exc, traceback)

    @staticmethod
    def signal_handler(sig, frame):
        """Handles the interrupt signal.

        Args:
            sig(int): Signal number of signal being passed to cli.
            frame(class):It represents execution frame. For more see
                https://docs.python.org/3/reference/datamodel.html

        Returns:
            None
        """
        log.info(f'Got {sig} signal. Stopping running steps.')
        for _, runner in WorkflowRunner.runners.items():
            runner.stop_running_tasks()
        sys.exit(0)

    @staticmethod
    def process_secrets(wf, config=DotMap({})):
        """Checks whether the secrets defined for a step are available in the
        execution environment. When the environment variable `CI` is set to
        `true` and no environment variable is defined for a secret, the
        execution fails, otherwise (CI not set or CI=false), it prompts the
        user to enter the values for undefined secrets. The processing is
        completely skipped if config.dry_run=True.

        Args:
          wf(popper.parser.Workflow): Instance of the Workflow class.
          config.dry_run(bool): skip execution of this function
          config.skip_clone(bool): skip execution of this function
        Returns:
            None
        """
        if config.dry_run or config.skip_clone:
            return

        for _, a in wf.steps.items():
            for s in a.get('secrets', []):
                if s not in os.environ:
                    if os.environ.get('CI', '') == 'true':
                        log.fail(f'Secret {s} not defined')
                    else:
                        val = getpass.getpass(
                            f'Enter the value for {s} : ')
                        os.environ[s] = val

    @staticmethod
    def clone_repos(wf, config):
        """Clone steps that reference a repository.

        Args:
          wf(popper.parser.workflow): Instance of the Workflow class.
          config.dry_run(bool): True if workflow flag is being dry-run.
          config.skip_clone(bool): True if clonning step has to be skipped.
          config.wid(str): id of the workspace

        Returns:
            None
        """
        repo_cache = os.path.join(pu.setup_base_cache(), config.wid)

        cloned = set()
        infoed = False

        for _, a in wf.steps.items():
            uses = a['uses']
            if 'docker://' in uses or './' in uses or uses == 'sh':
                continue

            url, service, user, repo, step_dir, version = scm.parse(
                a['uses'])

            repo_dir = os.path.join(repo_cache, service, user, repo)

            a['repo_dir'] = repo_dir
            a['step_dir'] = step_dir

            if config.dry_run:
                continue

            if config.skip_clone:
                if not os.path.exists(repo_dir):
                    log.fail(f"Expecting folder '{repo_dir}' not found.")
                continue

            if not infoed:
                log.info('[popper] Cloning step repositories')
                infoed = True

            if f'{user}/{repo}' in cloned:
                continue

            log.info(f'[popper] - {url}/{user}/{repo}@{version}')
            scm.clone(url, user, repo, repo_dir, version)
            cloned.add(f'{user}/{repo}')

    def run(self, wf):
        """Run the given workflow.

        Args:
          wf(Workflow): workflow to be executed

        Returns:
            None
        """
        WorkflowRunner.process_secrets(wf, self.config)
        WorkflowRunner.clone_repos(wf, self.config)

        for _, step in wf.steps.items():
            log.debug(f'Executing step:\n{pu.prettystr(step)}')
            if step['uses'] == 'sh':
                e = self.get_step_runner('host').run(step)
            else:
                e = self.get_step_runner(self.config.engine_name).run(step)

            if e != 0 and e != 78:
                log.fail(f"Step '{step['name']}' failed !")

            log.info(f"Step '{step['name']}' ran successfully !")

            if e == 78:
                break

        log.info(f"Workflow finished successfully.")

    def get_step_runner(self, rtype):
        """Singleton factory of runners"""
        if rtype not in self.runners:
            self.runners[rtype] = self.runner_classes[rtype](self.config)
        return self.runners[rtype]


class StepRunner(object):
    """Base class for step runners, assumed to be singletons."""
    def __init__(self, config):
        self.config = config

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, traceback):
        pass

    @staticmethod
    def prepare_environment(step, env={}):
        """Prepare environment variables for a step, which includes those in
        the 'env' and 'secrets' attributes.

        Args:
          step(dict): step information
          env(dict): optional environment to include in returned environment

        Returns:
          dict: key-value map of environment variables.
        """
        step_env = step.get('env', {}).copy()
        for s in step.get('secrets', []):
            step_env.update({s: os.environ[s]})
        step_env.update(env)
        return step_env

    def stop_running_tasks(self):
        raise NotImplementedError("Needs implementation in derived classes.")

    def run(self, step):
        raise NotImplementedError("Needs implementation in derived classes.")



import docker
import os

from popper import utils as pu
from popper import scm
from popper.cli import log
from popper.runner import StepRunner as StepRunner


class DockerRunner(StepRunner):
    """Runs steps in docker."""
    d = None

    # hold references to spawned containers
    spawned_containers = []

    def __init__(self, config):
        super(DockerRunner, self).__init__(config)

        try:
            DockerRunner.d = docker.from_env()
            DockerRunner.d.version()
        except Exception as e:
            log.debug(f'Docker error: {e}')
            log.fail(f'Unable to connect to the docker daemon.')

        log.debug(f'Docker info: {pu.prettystr(DockerRunner.d.info())}')

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if DockerRunner.d:
            DockerRunner.d.close()
        return True

    def run(self, step):
        """Execute the given step in docker."""
        cid = pu.sanitized_name(step['name'], self.config.wid)

        build, img, dockerfile = DockerRunner.get_build_info(
            step, self.config.workspace_dir, self.config.workspace_sha)

        container = DockerRunner.find_container(cid)

        if container and not self.config.reuse and not self.config.dry_run:
            container.remove(force=True)

        # build or pull
        if build:
            DockerRunner.docker_build(step, img, dockerfile,
                                      self.config.dry_run)
        elif not self.config.skip_pull and not step.get('skip_pull', False):
            DockerRunner.docker_pull(step, img, self.config.dry_run)

        msg = f'{img} {step.get("runs", "")} {step.get("args", "")}'
        log.info(f'[{step["name"]}] docker create {msg}')

        if not self.config.dry_run:
            engine_config = {
                "image": img,
                "command": step.get('args', None),
                "name": cid,
                "volumes": [
                    f'{self.config.workspace_dir}:/workspace',
                    '/var/run/docker.sock:/var/run/docker.sock'
                ],
                "working_dir": '/workspace',
                "environment": StepRunner.prepare_environment(step),
                "entrypoint": step.get('runs', None),
                "detach": True
            }

            if self.config.engine_options:
                DockerRunner.update_engine_config(engine_config,
                                                  self.config.engine_options)
            log.debug(f'Engine configuration: {pu.prettystr(engine_config)}\n')

            container = DockerRunner.d.containers.create(**engine_config)

        log.info(f'[{step["name"]}] docker start')

        if self.config.dry_run:
            return 0

        DockerRunner.spawned_containers.append(container)

        container.start()
        cout = container.logs(stream=True)
        for line in cout:
            log.step_info(pu.decode(line).strip('\n'))

        e = container.wait()['StatusCode']

        return e

    def stop_running_tasks(self):
        for c in DockerRunner.spawned_containers:
            log.info(f'Stopping container {c.name}')
            c.stop()

    @staticmethod
    def get_build_info(step, workspace_dir, workspace_sha):
        """Parses the `uses` attribute and returns build information needed.

        Args:
            step(dict): dict with step data

        Returns:
            (str, str, str): 'pull' or 'build', image ref, path to Dockerfile
        """
        build = True
        img = None
        build_source = None

        if 'docker://' in step['uses']:
            img = step['uses'].replace('docker://', '')
            if ':' not in img:
                img += ":latest"
            build = False
        elif './' in step['uses']:
            img = f'{pu.sanitized_name(step["name"], "step")}:{workspace_sha}'
            build_source = os.path.join(workspace_dir, step['uses'])
        else:
            _, _, user, repo, _, version = scm.parse(step['uses'])
            img = f'{user}/{repo}:{version}'
            build_source = os.path.join(step['repo_dir'], step['step_dir'])

        return (build, img.lower(), build_source)

    @staticmethod
    def find_container(cid):
        """Check whether the container exists."""
        containers = DockerRunner.d.containers.list(
            all=True, filters={'name': cid})

        filtered_containers = [c for c in containers if c.name == cid]

        if len(filtered_containers):
            return filtered_containers[0]

        return None

    @staticmethod
    def docker_image_exists(img):
        """Check whether a docker image exists for a step not.

        Args:
          img(str): The image to check for.

        Returns:
          bool: Whether the image exists or not.
        """
        images = DockerRunner.d.images.list(all=True)
        filtered_images = [i for i in images if img in i.tags]
        if filtered_images:
            return True
        return False

    @staticmethod
    def update_engine_config(engine_conf, update_with):
        engine_conf["volumes"] = [*engine_conf["volumes"],
                                  *update_with.get('volumes', list())]
        for k, v in update_with.get('environment', dict()).items():
            engine_conf["environment"].update({k: v})

        for k, v in update_with.items():
            if k not in engine_conf.keys():
                engine_conf[k] = update_with[k]

    @staticmethod
    def docker_pull(step, img, dry_run):
        """Pull an image from a container registry.

        Args:
          img(str): The image reference to pull.

        Returns:
            None
        """
        log.info(f'[{step["name"]}] docker pull {img}')
        if dry_run:
            return
        DockerRunner.d.images.pull(repository=img)

    @staticmethod
    def docker_build(step, tag, path, dry_run):
        """Build a docker image from a Dockerfile.

        Args:
          tag(str): The name of the image to build.
          path(str): The path to the Dockerfile.

        Returns:
            None
        """
        log.info(f'[{step["name"]}] docker build -t {tag} {path}')
        if dry_run:
            return
        DockerRunner.d.images.build(path=path, tag=tag, rm=True, pull=True)



import click

from popper.cli import pass_context, log
from popper.parser import Workflow


@click.option(
    '-f',
    '--wfile',
    help='File containing the definition of the workflow.',
    required=True
)
@click.option(
    '--skip',
    help=('Skip the list of steps specified.'),
    required=False,
    default=list(),
    multiple=True
)
@click.option(
    '--colors',
    help='Use colors in the graph.',
    required=False,
    is_flag=True,
)
@click.command('dot', short_help='Generate a graph in the .dot format.')
@pass_context
def cli(ctx, wfile, skip, colors):
    """Creates a graph in the .dot format representing the workflow.
    """

    # Args:
    #   ctx(Popper.cli.context): For process inter-command communication
    #         context is used.For reference visit
    #         https://click.palletsprojects.com/en/7.x/commands
    #   wfile(str): Name of the file containing definition of workflow.
    #   skip(tuple): List of steps that are to be skipped.
    #   colors(bool): Flag for colors.

    # Returns:
    #     None

    def add_to_graph(dot_str, wf, parent, children, node_attrs, stage_edges):
        """Recursively goes over the children ("next" attribute) of the given
        parent, adding an edge from parent to children

        Args:
          dot_str(str): The intermediate string to which further nodes are
                        to be added.
          wf(popper.parser.workflow): Instance of the workflow class.
          parent(str): Step Identifier.
          children(list/set): The node that is to be attached as a children.
          node_attrs(str): These are the attributes of the node of the graph.
          stage_edges(set): Intermediate sets containing the nodes and edges.

        Returns:
          str: The string containing nodes and their description.

        """
        for n in children:
            edge = f'  "{parent}" -> "{n}";\n'
            if edge in stage_edges:
                continue
            dot_str += edge + f'  "{n}" [{node_attrs}];\n'

            stage_edges.add(edge)

            for M in wf.steps[n].get('next', []):
                dot_str = add_to_graph(dot_str, wf, n, [M],
                                       node_attrs, stage_edges)
        return dot_str

    wf = Workflow.new(wfile)
    wf.parse()
    wf = Workflow.skip_steps(wf, skip)
    wf.check_for_unreachable_steps()

    node_attrs = (
        'shape=box, style="filled{}", fillcolor=transparent{}'
    )
    wf_attr = node_attrs.format(',rounded', ',color=red' if colors else '')
    act_attr = node_attrs.format('', ',color=cyan' if colors else '')
    dot_str = add_to_graph("", wf, wf.name, wf.root, act_attr, set())
    dot_str += f'  "{wf.name}" [{wf_attr}];\n'
    log.info(
        "digraph G { graph [bgcolor=transparent];\n" + dot_str + "}\n"
    )



from setuptools import setup

version = {}
with open('popper/__init__.py') as f:
    exec(f.read(), version)

setup(
    name='popper',
    version=version['__version__'],
    author='The Popper Development Team',
    author_email='ivo@cs.ucsc.edu',
    url='http://falsifiable.us',
    description='Popper CLI tool to generate reproducible papers.',
    packages=['popper', 'popper.commands'],
    include_package_data=True,
    install_requires=[
        'dotmap',
        'python-vagrant',
        'GitPython',
        'spython',
        'click',
        'pyhcl',
        'pyyaml',
        'docker'
    ],
    entry_points='''
        [console_scripts]
        popper=popper.cli:cli
    ''',
    project_urls={
        'Documentation': 'https://popper.readthedocs.io',
        'Say Thanks!': 'http://gitter.im/falsifiable-us/popper',
        'Source': 'https://github.com/systemslab/popper/',
        'Tracker': 'https://github.com/systemslab/popper/',
    },
    keywords='popper reproducibility automation continuous-integration',
    license='MIT',
)



import difflib
import os
import signal
import sys

import click
from click.exceptions import ClickException

import popper.log as logging

from popper import __version__

log = logging.setup_logging()
popper_version = __version__
cmd_folder = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                          'commands'))


class Context(object):
    """Context Class.For more details refer
    https://click.palletsprojects.com/en/7.x/commands/#custom-multi-commands .
    """
    pass


pass_context = click.make_pass_decorator(Context, ensure=True)


class PopperCLI(click.MultiCommand):
    """Provides CLI interface for Popper."""

    def list_commands(self, ctx):
        """Returns the list of available commands in sorted order.

        Args:
          ctx(popper.cli.context): For process inter-command communication.
          For reference visit https://click.palletsprojects.com/en/7.x

        Returns:
          list: Returns the list of available commands.
        """
        rv = []
        for filename in os.listdir(cmd_folder):
            if filename.endswith('.py') and filename.startswith('cmd_'):
                rv.append(filename[4:-3])
        rv.sort()
        return rv

    def get_command(self, ctx, name):
        """Imports the command if available in commmands list and provides with
        most similar commands if the command is not present in the list.

        Args:
          name(str): The name of the command.

        Returns:
          click.core.Command: It is a new command and uses the decorated
            function as callback.For reference visit
            https://click.palletsprojects.com/en/7.x/api/#decorators .
        """
        try:
            if sys.version_info[0] == 2:
                name = name.encode('ascii', 'replace')
            mod = __import__('popper.commands.cmd_' + name,
                             None, None, ['cli'])
        except ImportError as e:
            commands = self.list_commands(ctx)
            most_similar_commands = ", ".join(
                difflib.get_close_matches(
                    name, commands, 3, 0.3))
            message = ""
            if len(most_similar_commands) != 0:
                message = "\n\nThe most similar commands are: " \
                    + most_similar_commands
            raise ClickException(
                "Command '" + name + "' doesn't exist.\n"
                "Type 'popper --help' for more.\n" + message + '\n' + str(e)
            )
        return mod.cli


@click.command(cls=PopperCLI)
@pass_context
def cli(ctx):
    """Popper command line interface."""
    from popper.runner import WorkflowRunner
    signal.signal(signal.SIGINT, WorkflowRunner.signal_handler)
    signal.signal(signal.SIGUSR1, WorkflowRunner.signal_handler)



import os
import unittest

import utils as testutils

from popper.parser import YMLWorkflow
from popper.runner import WorkflowRunner

from popper.cli import log as log


class TestDockerRunner(unittest.TestCase):

    def setUp(self):
        log.setLevel('CRITICAL')

    @unittest.skipIf(os.environ['ENGINE'] != 'docker', 'ENGINE != docker')
    def test_basic_run(self):
        repo = testutils.mk_repo()

        with WorkflowRunner(workspace_dir=repo.working_dir) as r:
            wf = YMLWorkflow("""
            version: '1'
            steps:
            - uses: 'popperized/bin/sh@master'
              runs: [cat]
              args: README.md
            """)
            wf.parse()
            r.run(wf)

            wf = YMLWorkflow("""
            version: '1'
            steps:
            - uses: 'docker://alpine:3.9'
              runs: ['sh', '-c', 'echo $FOO > hello.txt ; pwd']
              env: {
                  FOO: bar
              }
            """)
            wf.parse()
            r.run(wf)
            with open(os.path.join(repo.working_dir, 'hello.txt'), 'r') as f:
                self.assertEqual(f.read(), 'bar\n')

            wf = YMLWorkflow("""
            version: '1'
            steps:
            - uses: 'docker://alpine:3.9'
              runs: 'nocommandisnamedlikethis'
            """)
            wf.parse()
            self.assertRaises(Exception, r.run, wf)

        repo.close()

    @unittest.skipIf(os.environ['ENGINE'] != 'docker', 'ENGINE != docker')
    def test_local_dockerfile(self):
        pass

    def test_skip_pull(self):
        pass
